@@ -1,4 +1,3 @@
-
 import warnings
 warnings.filterwarnings("ignore", "apt API not stable yet", FutureWarning)
 import apt
@@ -8,41 +7,128 @@ import os.path
 import re
 import logging
 import string
+import time
+import gettext
+import datetime
+import threading
 from subprocess import Popen, PIPE
 
 from gettext import gettext as _
 from DistUpgradeConfigParser import DistUpgradeConfig
 from DistUpgradeView import FuzzyTimeToStr
 
-# FIXME: we need this only for the later "isinstance()" check
-#        this should probably be solved in some different way
-from DistUpgradeViewText import DistUpgradeViewText
-from DistUpgradeViewNonInteractive import DistUpgradeViewNonInteractive
 
 class CacheException(Exception):
     pass
 class CacheExceptionLockingFailed(CacheException):
     pass
+class CacheExceptionDpkgInterrupted(CacheException):
+    pass
 
 class MyCache(apt.Cache):
+    ReInstReq = 1
+    HoldReInstReq = 3
+
     # init
     def __init__(self, config, view, progress=None, lock=True):
         apt.Cache.__init__(self, progress)
         self.to_install = []
         self.to_remove = []
         self.view = view
+        self.lock = False
         self.config = config
         self.metapkgs = self.config.getlist("Distro","MetaPkgs")
         # acquire lock
+        self._listsLock = -1
         if lock:
             try:
                 apt_pkg.PkgSystemLock()
+                self.lockListsDir()
                 self.lock = True
             except SystemError, e:
+                # checking for this is ok, its not translatable
+                if "dpkg was interrupted" in str(e):
+                    raise CacheExceptionDpkgInterrupted, e
                 raise CacheExceptionLockingFailed, e
         # a list of regexp that are not allowed to be removed
         self.removal_blacklist = config.getListFromFile("Distro","RemovalBlacklistFile")
         self.uname = Popen(["uname","-r"],stdout=PIPE).communicate()[0].strip()
+        self._initAptLog()
+
+    @property
+    def reqReinstallPkgs(self):
+        " return the packages not downloadable packages in reqreinst state "
+        reqreinst = set()
+        for pkg in self:
+            if (not pkg.candidateDownloadable and 
+                (pkg._pkg.InstState == self.ReInstReq or
+                 pkg._pkg.InstState == self.HoldReInstReq)):
+                reqreinst.add(pkg.name)
+        return reqreinst
+
+    def fixReqReinst(self, view):
+        " check for reqreinst state and offer to fix it "
+        reqreinst = self.reqReinstallPkgs
+        if len(reqreinst) > 0:
+            header = gettext.ngettext("Remove package in bad state",
+                                      "Remove packages in bad state", 
+                                      len(reqreinst))
+            summary = gettext.ngettext("The package '%s' is in a inconsistent "
+                                       "state and needs to be reinstalled, but "
+                                       "no archive can be found for it. "
+                                       "Do you want to remove this package "
+                                       "now to continue?",
+                                       "The packages '%s' is in a inconsistent "
+                                       "state and needs to be reinstalled, but "
+                                       "no archive can be found for it. Do you "
+                                       "want to remove this package now to "
+                                       "continue?",
+                                       len(reqreinst)) % ", ".join(reqreinst)
+            if view.askYesNoQuestion(header, summary):
+                self.releaseLock()
+                cmd = ["dpkg","--remove","--force-remove-reinstreq"] + list(reqreinst)
+                view.getTerminal().call(cmd)
+                self.getLock()
+                return True
+        return False
+
+    # logging stuff
+    def _initAptLog(self):
+        " init loging, create log file"
+        logdir = self.config.getWithDefault("Files","LogDir",
+                                            "/var/log/dist-upgrade")
+        apt_pkg.Config.Set("Dir::Log",logdir)
+        apt_pkg.Config.Set("Dir::Log::Terminal","apt-term.log")
+        self.logfd = os.open(os.path.join(logdir,"apt.log"),
+                             os.O_RDWR|os.O_CREAT|os.O_APPEND|os.O_SYNC, 0644)
+        os.write(self.logfd, "Log time: %s\n" % datetime.datetime.now())
+        # turn on debuging in the cache
+        apt_pkg.Config.Set("Debug::pkgProblemResolver","true")
+        apt_pkg.Config.Set("Debug::pkgDepCache::AutoInstall","true")
+    def _startAptResolverLog(self):
+        if hasattr(self, "old_stdout"):
+            os.close(self.old_stdout)
+            os.close(self.old_stderr)
+        self.old_stdout = os.dup(1)
+        self.old_stderr = os.dup(2)
+        os.dup2(self.logfd, 1)
+        os.dup2(self.logfd, 2)
+    def _stopAptResolverLog(self):
+        os.fsync(1)
+        os.fsync(2)
+        os.dup2(self.old_stdout, 1)
+        os.dup2(self.old_stderr, 2)
+    # use this decorator instead of the _start/_stop stuff directly
+    # FIXME: this should probably be a decorator class where all
+    #        logging is moved into?
+    def withResolverLog(f):
+        " decorator to ensure that the apt output is logged "
+        def wrapper(*args, **kwargs):
+            args[0]._startAptResolverLog()
+            res = f(*args, **kwargs)
+            args[0]._stopAptResolverLog()
+            return res
+        return wrapper
 
     # properties
     @property
@@ -62,13 +148,32 @@ class MyCache(apt.Cache):
         return self._depcache.BrokenCount > 0
 
     # methods
+    def lockListsDir(self):
+        name = apt_pkg.Config.FindDir("Dir::State::Lists") + "lock"
+        self._listsLock = apt_pkg.GetLock(name)
+        if self._listsLock < 0:
+            e = "Can not lock '%s' " % name
+            raise CacheExceptionLockingFailed, e
+    def unlockListsDir(self):
+        if self._listsLock > 0:
+            os.close(self._listsLock)
+            self._listsLock = -1
+    def update(self, fprogress=None):
+        """
+        our own update implementation is required because we keep the lists
+        dir lock
+        """
+        self.unlockListsDir()
+        apt.Cache.update(self, fprogress)
+        self.lockListsDir()
+
     def commit(self, fprogress, iprogress):
         logging.info("cache.commit()")
         if self.lock:
             self.releaseLock()
         apt.Cache.commit(self, fprogress, iprogress)
 
-    def releaseLock(self):
+    def releaseLock(self, pkgSystemOnly=True):
         if self.lock:
             try:
                 apt_pkg.PkgSystemUnLock()
@@ -76,6 +181,14 @@ class MyCache(apt.Cache):
             except SystemError, e:
                 logging.debug("failed to SystemUnLock() (%s) " % e)
 
+    def getLock(self, pkgSystemOnly=True):
+        if not self.lock:
+            try:
+                apt_pkg.PkgSystemLock()
+                self.lock = True
+            except SystemError, e:
+                logging.debug("failed to SystemLock() (%s) " % e)
+
     def downloadable(self, pkg, useCandidate=True):
         " check if the given pkg can be downloaded "
         if useCandidate:
@@ -83,6 +196,7 @@ class MyCache(apt.Cache):
         else:
             ver = pkg._pkg.CurrentVer
         if ver == None:
+            logging.warning("no version information for '%s' (useCandidate=%s)" % (pkg.name, useCandidate))
             return False
         return ver.Downloadable
     
@@ -167,19 +281,23 @@ class MyCache(apt.Cache):
                                       self[key].markedInstall):
                 for pkgname in self.config.getlist(key,"KeepInstalledPkgs"):
                     keepInstalled(self, pkgname, "%s KeepInstalledPkgs rule" % key)
-        # now the keepInstalledSection code
-        for section in self.config.getlist("Distro","KeepInstalledSection"):
-            for pkg in self:
-                if pkg.markedDelete and pkg.section == section:
-                    keepInstalled(self, pkg.name, "Distro KeepInstalledSection rule: %s" % section)
-        # the the per-metapkg rules
-        for key in self.metapkgs:
-            if self.has_key(key) and (self[key].isInstalled or
-                                      self[key].markedInstall):
-                for section in self.config.getlist(key,"KeepInstalledSection"):
-                    for pkg in self:
-                        if pkg.markedDelete and pkg.section == section:
-                            keepInstalled(self, pkg.name, "%s KeepInstalledSection rule: %s" % (key, section))
+
+        # only enforce section if we have a network. Otherwise we run
+        # into CD upgrade issues for installed language packs etc
+        if bool(self.config.get("Options","withNetwork")):
+            logging.debug("Running KeepInstalledSection rules")
+            # now the keepInstalledSection code
+            for section in self.config.getlist("Distro","KeepInstalledSection"):
+                for pkg in self:
+                    if pkg.markedDelete and pkg.section == section:
+                        keepInstalled(self, pkg.name, "Distro KeepInstalledSection rule: %s" % section)
+            for key in self.metapkgs:
+                if self.has_key(key) and (self[key].isInstalled or
+                                          self[key].markedInstall):
+                    for section in self.config.getlist(key,"KeepInstalledSection"):
+                        for pkg in self:
+                            if pkg.markedDelete and pkg.section == section:
+                                keepInstalled(self, pkg.name, "%s KeepInstalledSection rule: %s" % (key, section))
         
 
     def postUpgradeRule(self):
@@ -197,10 +315,69 @@ class MyCache(apt.Cache):
                         action(pkg, "%s PostUpgrade%s rule" % (key, rule))
 
         # get the distro-specific quirks handler and run it
-        quirksFuncName = "%sQuirks" % self.config.get("Sources","To")
-        func = getattr(self, quirksFuncName, None)
-        if func is not None:
-            func()
+        for quirksFuncName in ("%sQuirks" % self.config.get("Sources","To"),
+                               "from_%sQuirks" % self.config.get("Sources","From")):
+            func = getattr(self, quirksFuncName, None)
+            if func is not None:
+                func()
+
+    def from_dapperQuirks(self):
+        self.hardyQuirks()
+        self.gutsyQuirks()
+        self.feistyQuirks()
+        self.edgyQuirks()
+
+    def _checkAndRemoveEvms(self):
+        " check if evms is in use and if not, remove it "
+        logging.debug("running _checkAndRemoveEvms")
+        for line in open("/proc/mounts"):
+            line = line.strip()
+            if line == '' or line.startswith("#"):
+                continue
+            try:
+                (device, mount_point, fstype, options, a, b) = line.split()
+            except Exception, e:
+                logging.error("can't parse line '%s'" % line)
+                continue
+            if "evms" in device:
+                logging.debug("found evms device in line '%s', skipping " % line)
+                return False
+        # if not in use, nuke it
+        for pkg in ["evms","libevms-2.5","libevms-dev",
+                    "evms-ncurses", "evms-ha",
+                    "evms-bootdebug",
+                    "evms-gui", "evms-cli",
+                    "linux-patch-evms"]:
+            if self.has_key(pkg) and self[pkg].isInstalled:
+                self[pkg].markDelete()
+        return True
+
+
+    def hardyQuirks(self):
+        """ 
+        this function works around quirks in the 
+        {dapper,gutsy}->hardy upgrade 
+        """
+        logging.debug("running hardyQuirks handler")
+        # deal with gnome-translator and help apt with the breaks
+        if (self.has_key("nautilus") and
+            self["nautilus"].isInstalled and
+            not self["nautilus"].markedUpgrade):
+            # uninstallable and gutsy apt is unhappy about this
+            # breaks because it wants to upgrade it and gives up
+            # if it can't
+            for broken in ("link-monitor-applet"):
+                if self.has_key(broken) and self[broken].isInstalled:
+                    self[broken].markDelete()
+            self["nautilus"].markInstall()
+        # evms gives problems, remove it if it is not in use
+        self._checkAndRemoveEvms()
+        # give the language-support-* packages a extra kick
+        for pkg in self:
+            if (pkg.name.startswith("language-support-") and
+                pkg.isInstalled and
+                not pkg.markedUpgrade):
+                self.markInstall(pkg.name,"extra language-support- kick")
 
     def gutsyQuirks(self):
         """ this function works around quirks in the feisty->gutsy upgrade """
@@ -208,7 +385,9 @@ class MyCache(apt.Cache):
         # lowlatency kernel flavour vanished from feisty->gutsy
         try:
             (version, build, flavour) = self.uname.split("-")
-            if flavour == 'lowlatency':
+            if (flavour == 'lowlatency' or 
+                flavour == '686' or
+                flavour == 'k7'):
                 kernel = "linux-image-generic"
                 if not (self[kernel].isInstalled or self[kernel].markedInstall):
                     logging.debug("Selecting new kernel '%s'" % kernel)
@@ -254,7 +433,9 @@ class MyCache(apt.Cache):
                 pkg.isInstalled and
                 not pkg.markedUpgrade):
                 basepkg = "python-"+pkg.name[len("python2.4-"):]
-                if (self.has_key(basepkg) and not self[basepkg].markedInstall):
+                if (self.has_key(basepkg) and 
+                    self[basepkg].candidateDownloadable and
+                    not self[basepkg].markedInstall):
                     try:
                         self.markInstall(basepkg,
                                          "python2.4->python upgrade rule")
@@ -272,7 +453,7 @@ class MyCache(apt.Cache):
         # deal with held-backs that are unneeded
         for pkgname in ["hpijs", "bzr", "tomboy"]:
             if (self.has_key(pkgname) and self[pkgname].isInstalled and
-                not self[pkgname].markedUpgrade):
+                self[pkgname].isUpgradable and not self[pkgname].markedUpgrade):
                 try:
                     self.markInstall(pkgname,"%s quirk upgrade rule" % pkgname)
                 except SystemError, e:
@@ -387,17 +568,19 @@ class MyCache(apt.Cache):
                 pkg.priority in need):
                 self.markInstall(pkg.name, "priority in required set '%s' but not scheduled for install" % need)
 
-    def distUpgrade(self, view, serverMode, logfd):
-        def _restore_fds(stdout, stderr):
-            os.dup2(stdout, 1)
-            os.dup2(stderr, 2)
-        text_mode = (isinstance(view, DistUpgradeViewText) or
-                     isinstance(view, DistUpgradeViewNonInteractive))
-        if text_mode:
-            old_stdout = os.dup(1)
-            old_stderr = os.dup(2)
-            os.dup2(logfd, 1)
-            os.dup2(logfd, 2)
+    # FIXME: make this a decorator (just like the withResolverLog())
+    def updateGUI(self, view, lock):
+        while lock.locked():
+            view.processEvents()
+            time.sleep(0.01)
+
+    @withResolverLog
+    def distUpgrade(self, view, serverMode, partialUpgrade):
+        # keep the GUI alive
+        lock = threading.Lock()
+        lock.acquire()
+        t = threading.Thread(target=self.updateGUI, args=(self.view, lock,))
+        t.start()
         try:
             # upgrade (and make sure this way that the cache is ok)
             self.upgrade(True)
@@ -424,20 +607,42 @@ class MyCache(apt.Cache):
 
             # see if it all makes sense
             if not self._verifyChanges():
-                if text_mode: _restore_fds(old_stdout, old_stderr)
                 raise SystemError, _("A essential package would have to be removed")
         except SystemError, e:
+            # this should go into a finally: line, see below for the 
+            # rational why it dosn't 
+            lock.release()
+            t.join()
             # FIXME: change the text to something more useful
-            view.error(_("Could not calculate the upgrade"),
-                       _("A unresolvable problem occurred while "
-                         "calculating the upgrade.\n\n"
-                         "Please report this bug against the 'update-manager' "
-                         "package and include the files in /var/log/dist-upgrade/ "
-                         "in the bugreport."))
+            details =  _("A unresolvable problem occurred while "
+                         "calculating the upgrade.\n\n "
+                         "This can be caused by:\n"
+                         " * Upgrading to a pre-release version of Ubuntu\n"
+                         " * Running the current pre-release version of Ubuntu\n"
+                         " * Unofficial software packages not provided by Ubuntu\n"
+                         "\n")
+            # we never have partialUpgrades (including removes) on a stable system
+            # with only ubuntu sources so we do not recommend reporting a bug
+            if partialUpgrade:
+                details += _("This is most likely a transient problem, "
+                             "please try again later.")
+            else:
+                details += _("If none of this applies, then please report this bug against "
+                             "the 'update-manager' package and include the files in "
+                             "/var/log/dist-upgrade/ in the bugreport.")
+            # make the text available again
+            self._stopAptResolverLog()
+            view.error(_("Could not calculate the upgrade"), details)
+            self._startAptResolverLog()            
             logging.error("Dist-upgrade failed: '%s'", e)
-            if text_mode: _restore_fds(old_stdout, old_stderr)
             return False
-
+        # would be nice to be able to use finally: here, but we need
+        # to run on python2.4 too 
+        #finally:
+        # wait for the gui-update thread to exit
+        lock.release()
+        t.join()
+        
         # check the trust of the packages that are going to change
         untrusted = []
         for pkg in self.getChanges():
@@ -449,8 +654,10 @@ class MyCache(apt.Cache):
                     # version is lower than installed one
                     if apt_pkg.VersionCompare(ver.VerStr, pkg.installedVersion) < 0:
                         for (verFileIter,index) in ver.FileList:
-                            if not origin.trusted:
+                            indexfile = pkg._list.FindIndex(verFileIter)
+                            if indexfile and not indexfile.IsTrusted:
                                 untrusted.append(pkg.name)
+                                break
                 continue
             origins = pkg.candidateOrigin
             trusted = False
@@ -473,9 +680,7 @@ class MyCache(apt.Cache):
                          "You may want to try again later. See below for a "
                          "list of unauthenticated packages."),
                        "\n".join(untrusted))
-            if text_mode: _restore_fds(old_stdout, old_stderr)
             return False
-        if text_mode: _restore_fds(old_stdout, old_stderr)
         return True
 
     def _verifyChanges(self):
@@ -498,7 +703,9 @@ class MyCache(apt.Cache):
         tasks = {}
         installed_tasks = set()
         for pkg in self:
-            pkg._lookupRecord()
+            if not pkg._lookupRecord():
+                logging.debug("no PkgRecord found for '%s', skipping " % pkg.name)
+                continue
             for line in pkg._records.Record.split("\n"):
                 if line.startswith("Task:"):
                     for task in (line[len("Task:"):]).split(","):
@@ -597,7 +804,9 @@ class MyCache(apt.Cache):
                 return True
         return False
 
-    def _tryMarkObsoleteForRemoval(self, pkgname, remove_candidates, foreign_pkgs):
+    @withResolverLog
+    def tryMarkObsoleteForRemoval(self, pkgname, remove_candidates, foreign_pkgs):
+        logging.debug("tryMarkObsoleteForRemoval(): %s" % pkgname)
         # sanity check, first see if it looks like a runing kernel pkg
         if pkgname.endswith(self.uname):
             logging.debug("skipping runing kernel pkg '%s'" % pkgname)
@@ -605,17 +814,26 @@ class MyCache(apt.Cache):
         if self._inRemovalBlacklist(pkgname):
             logging.debug("skipping '%s' (in removalBlacklist)" % pkgname)
             return False
+        # if we don't have the package anyway, we are fine (this can
+        # happen when forced_obsoletes are specified in the config file)
+        if not self.has_key(pkgname):
+            #logging.debug("package '%s' not in cache" % pkgname)
+            return True
+        # check if we want to purge 
+        purge = bool(self.config.getWithDefault("Distro","PurgeObsoletes",False))
         # this is a delete candidate, only actually delete,
         # if it dosn't remove other packages depending on it
         # that are not obsolete as well
         actiongroup = apt_pkg.GetPkgActionGroup(self._depcache)
         self.create_snapshot()
         try:
-            self[pkgname].markDelete()
+            self[pkgname].markDelete(purge=purge)
+            logging.debug("marking '%s' for removal" % pkgname)
             for pkg in self.getChanges():
                 if (pkg.name not in remove_candidates or 
                       pkg.name in foreign_pkgs or 
                       self._inRemovalBlacklist(pkg.name)):
+                    logging.debug("package '%s' has unwanted removals, skipping" % pkgname)
                     self.restore_snapshot()
                     return False
         except (SystemError,KeyError),e:
@@ -628,11 +846,22 @@ class MyCache(apt.Cache):
         " get all package names that are not downloadable "
         obsolete_pkgs =set()        
         for pkg in self:
-            if pkg.isInstalled:
-                if not self.downloadable(pkg):
+            if pkg.isInstalled: 
+                # check if any version is downloadable. we need to check
+                # for older ones too, because there might be
+                # cases where e.g. firefox in gutsy-updates is newer
+                # than hardy
+                if not self.anyVersionDownloadable(pkg):
                     obsolete_pkgs.add(pkg.name)
         return obsolete_pkgs
 
+    def anyVersionDownloadable(self, pkg):
+        " helper that checks if any of the version of pkg is downloadable "
+        for ver in pkg._pkg.VersionList:
+            if ver.Downloadable:
+                return True
+        return False
+
     def _getUnusedDependencies(self):
         " get all package names that are not downloadable "
         unused_dependencies =set()        
