@@ -1,6 +1,6 @@
 # UpdateManager.py 
 #  
-#  Copyright (c) 2004-2006 Canonical
+#  Copyright (c) 2004-2008 Canonical
 #                2004 Michiel Sikkes
 #                2005 Martin Willemoes Hansen
 #  
@@ -44,6 +44,7 @@ import string
 import sys
 import os
 import os.path
+import stat
 import re
 import locale
 import tempfile
@@ -51,6 +52,7 @@ import pango
 import subprocess
 import pwd
 import urllib2
+import httplib
 import time
 import thread
 import xml.sax.saxutils
@@ -90,9 +92,18 @@ CHANGELOGS_URI="http://changelogs.ubuntu.com/changelogs/pool/%s/%s/%s/%s_%s/chan
 class MyCache(apt.Cache):
     def __init__(self, progress, rootdir=None):
         apt.Cache.__init__(self, progress, rootdir)
+        # raise if we have packages in reqreinst state
+        # and let the caller deal with that (runs partial upgrade)
+        assert len(self.reqReinstallPkgs) == 0
+        # init the regular cache
         self._initDepCache()
-        assert self._depcache.BrokenCount == 0 and self._depcache.DelCount == 0
         self.all_changes = {}
+        # on broken packages, try to fix via saveDistUpgrade()
+        if self._depcache.BrokenCount > 0:
+            self.saveDistUpgrade()
+        assert (self._depcache.BrokenCount == 0 and 
+                self._depcache.DelCount == 0)
+
     def _initDepCache(self):
         #apt_pkg.Config.Set("Debug::pkgPolicy","1")
         #self.depcache = apt_pkg.GetDepCache(self.cache)
@@ -241,7 +252,7 @@ class MyCache(apt.Cache):
                 self.all_changes[name] = [_("The list of changes is not "
                                             "available yet.\nPlease try again "
                                             "later."), srcpkg]
-        except IOError:
+        except IOError, httplib.BadStatusLine:
             if lock.locked():
                 self.all_changes[name] = [_("Failed to download the list "
                                             "of changes. \nPlease "
@@ -421,12 +432,13 @@ class UpdateManager(SimpleGladeApp):
           host = self.gconfclient.get_string("/system/http_proxy/host")
           port = self.gconfclient.get_int("/system/http_proxy/port")
           use_auth = self.gconfclient.get_bool("/system/http_proxy/use_authentication")
-          if use_auth:
-              auth_user = self.gconfclient.get_string("/system/http_proxy/authentication_user")
-              auth_pw = self.gconfclient.get_string("/system/http_proxy/authentication_password")
-              proxy = "http://%s:%s@%s:%s/" % (auth_user,auth_pw,host, port)
-          else:
-              proxy = "http://%s:%s/" % (host, port)
+          if host and port:
+              if use_auth:
+                  auth_user = self.gconfclient.get_string("/system/http_proxy/authentication_user")
+                  auth_pw = self.gconfclient.get_string("/system/http_proxy/authentication_password")
+                  proxy = "http://%s:%s@%s:%s/" % (auth_user,auth_pw,host, port)
+              else:
+                  proxy = "http://%s:%s/" % (host, port)
       if proxy:
           proxy_support = urllib2.ProxyHandler({"http":proxy})
           opener = urllib2.build_opener(proxy_support)
@@ -576,7 +588,7 @@ class UpdateManager(SimpleGladeApp):
                             lambda w,lock: lock.release(), lock)
         # wait for the dl-thread
         while lock.locked():
-          time.sleep(0.05)
+          time.sleep(0.01)
           while gtk.events_pending():
             gtk.main_iteration()
         # download finished (or canceld, or time-out)
@@ -648,12 +660,35 @@ class UpdateManager(SimpleGladeApp):
       except SystemError, e:
           print "requiredDownload could not be calculated: %s" % e
           self.label_downsize.set_markup(_("Unknown download size"))
-      
+
+  def _get_last_apt_get_update_text(self):
+      """
+      return a human readable string with the information when
+      the last apt-get update was run
+      """
+      if not os.path.exists("/var/lib/apt/periodic/update-success-stamp"):
+          return _("It is unknown when the package information was "
+                   "updated last. Please try clicking on the 'Check' "
+                   "button to update the information.")
+      # calculate when the last apt-get update (or similar operation)
+      # was performed
+      mtime = os.stat("/var/lib/apt/periodic/update-success-stamp")[stat.ST_MTIME]
+      ago_days = int( (time.time() - mtime) / (24*60*60))
+      ago_hours = int((time.time() - mtime) / (60*60) )
+      if ago_days > 0:
+          return _("The package information was last updated %s days ago.") % ago_days
+      elif ago_hours > 0:
+          return _("The package information was last updated %s hours ago.") % ago_hours
+      else:
+          return _("The package information was last updated less than one hour ago.")
+      return None
+
   def update_count(self):
       """activate or disable widgets and show dialog texts correspoding to
          the number of available updates"""
       self.refresh_updates_count()
       num_updates = self.cache.installCount
+      text_label_main = _("Software updates correct errors, eliminate security vulnerabilities and provide new features.")
       if num_updates == 0:
           text_header= "<big><b>%s</b></big>"  % _("Your system is up-to-date")
           text_download = ""
@@ -664,10 +699,12 @@ class UpdateManager(SimpleGladeApp):
           self.button_close.grab_default()
           self.textview_changes.get_buffer().set_text("")
           self.textview_descr.get_buffer().set_text("")
+          if self._get_last_apt_get_update_text() is not None:
+              text_label_main = self._get_last_apt_get_update_text()
       else:
           text_header = "<big><b>%s</b></big>" % \
-                        (gettext.ngettext("You can install %s update",
-                                          "You can install %s updates", 
+                        (gettext.ngettext("You can install %s update.",
+                                          "You can install %s updates.", 
                                           num_updates) % \
                                           num_updates)
           text_download = _("Download size: %s") % humanize_size(self.dl_size)
@@ -677,6 +714,7 @@ class UpdateManager(SimpleGladeApp):
           self.treeview_update.set_cursor(1)
       self.label_header.set_markup(text_header)
       self.label_downsize.set_markup(text_download)
+      self.label_main_details.set_text(text_label_main)
 
   def activate_details(self, expander, data):
     expanded = self.expander_details.get_expanded()
@@ -694,10 +732,17 @@ class UpdateManager(SimpleGladeApp):
       apt_pkg.PkgSystemUnLock()
     except SystemError:
       pass
-    cmd = ["gksu", "--desktop", "/usr/share/applications/update-manager.desktop", 
+    cmd = ["/usr/bin/gksu", 
+           "--desktop", "/usr/share/applications/update-manager.desktop", 
            "--", "/usr/sbin/synaptic", "--hide-main-window",  
            "--non-interactive", "--parent-window-id", "%s" % (id) ]
     if action == INSTALL:
+      # close when update was successful (its ok to use a Synaptic::
+      # option here, it will not get auto-saved, because synaptic does
+      # not save options in non-interactive mode)
+      cmd.append("-o")
+      cmd.append("Synaptic::closeZvt=true")
+      # custom progress strings
       cmd.append("--progress-str")
       cmd.append("%s" % _("Please wait, this can take some time."))
       cmd.append("--finish-str")
@@ -753,7 +798,8 @@ class UpdateManager(SimpleGladeApp):
       time.sleep(0.05)
     while gtk.events_pending():
       gtk.main_iteration()
-    self.label_cache_progress_title.set_label("<b><big>%s</big></b>" % _("Checking for updates"))
+    s = _("Reading package information")
+    self.label_cache_progress_title.set_label("<b><big>%s</big></b>" % s)
     self.fillstore()
 
     # Allow suspend after synaptic is finished
@@ -799,8 +845,12 @@ class UpdateManager(SimpleGladeApp):
   def save_state(self):
     """ save the state  (window-size for now) """
     (x,y) = self.window_main.get_size()
-    self.gconfclient.set_pair("/apps/update-manager/window_size",
-                              gconf.VALUE_INT, gconf.VALUE_INT, x, y)
+    try:
+        self.gconfclient.set_pair("/apps/update-manager/window_size",
+                                  gconf.VALUE_INT, gconf.VALUE_INT, x, y)
+    except gobject.GError, e:
+        print "Could not save the configuration to gconf: %s" % e
+        pass
 
   def restore_state(self):
     """ restore the state (window-size for now) """
@@ -872,16 +922,9 @@ class UpdateManager(SimpleGladeApp):
           name = xml.sax.saxutils.escape(pkg.name)
           summary = xml.sax.saxutils.escape(pkg.summary)
           contents = "<b>%s</b>\n<small>%s</small>" % (name, summary)
-          if pkg.installedVersion != None:
-              version =  _("From version %(old_version)s to %(new_version)s") %\
-                         {"old_version" : pkg.installedVersion,
-                          "new_version" : pkg.candidateVersion}
-          else:
-              version = _("Version %s") % pkg.candidateVersion
           #TRANSLATORS: the b stands for Bytes
           size = _("(Size: %s)") % humanize_size(pkg.packageSize)
-          contents = "%s\n<small>%s %s</small>" % (contents, version, size)
-
+          contents = "%s <small>%s</small>" % (contents, size)
           self.store.append([contents, pkg.name, pkg])
     self.update_count()
     self.setBusy(False)
@@ -953,6 +996,10 @@ class UpdateManager(SimpleGladeApp):
         else:
             self.cache = MyCache(progress)
     except AssertionError:
+        # if the cache could not be opened for some reason,
+        # let the release upgrader handle it, it deals
+        # a lot better with this
+        self.ask_run_partial_upgrade()
         # we assert a clean cache
         msg=("<big><b>%s</b></big>\n\n%s"% \
              (_("Software index is broken"),
@@ -990,6 +1037,9 @@ class UpdateManager(SimpleGladeApp):
     """ Check if all available updates can be installed and suggest
         to run a distribution upgrade if not """
     if self.list.distUpgradeWouldDelete > 0:
+        self.ask_run_partial_upgrade()
+
+  def ask_run_partial_upgrade(self):
       self.dialog_dist_upgrade.set_transient_for(self.window_main)
       res = self.dialog_dist_upgrade.run()
       self.dialog_dist_upgrade.hide()
@@ -998,6 +1048,7 @@ class UpdateManager(SimpleGladeApp):
                    "/usr/bin/gksu", "--desktop",
                    "/usr/share/applications/update-manager.desktop",
                    "--", "/usr/bin/update-manager", "--dist-upgrade")
+      return False
 
   def check_metarelease(self):
       " check for new meta-release information "
